"""
This script is intended to run single or multithreaded experiments.
"""
__license__ = "BSD-3"
__docformat__ = 'reStructuredText'
__author__ = "Jared Beard"

from importlib.resources import path
import sys
import os

current = os.path.dirname(os.path.realpath(__file__))
parent = os.path.dirname(current)
sys.path.append(parent)

from multiprocessing import Pool, Lock
import logging
import itertools
import pandas as pd
from reconfigurator.reconfigurator import compile_as_generator
from copy import deepcopy

import nestifydict as nd

from file_utils import *

class Experiment():
    """
    Perform trials for a given set of experiments. 

    Users should provide two parameter files.

    :param trial_config: (list(dict)/generator) Configurations for each trial, *default*: None
    :param expt_config: (dict) Experiment configuration file containing the following keys:
        - *default*: None
        - "experiment": (func) Reference to function under test
        - "n_trials": (int) Number of trials to run for each set of parameters, *default*: 1
        - "n_processes": (int) Number of processes to use, *default*: 1
        - "save_file": (str) file to save data, if none does not save, *default*: None
        - "clear_save": (bool) clears data from pickle before running experiment, *default*: False
    :param log_level: (str) Logging level, *default*: "WARNING"
    """
    def __init__(self, trial_config = None, expt_config : dict = None, log_level : str = "WARNING"):
        
        super(Experiment, self).__init__()
        
        log_levels = {"NOTSET": logging.NOTSET, "DEBUG": logging.DEBUG, "INFO": logging.INFO, "WARNING": logging.WARNING, "ERROR": logging.ERROR ,"CRITICAL": logging.CRITICAL}
        self._log_level = log_levels[log_level]
                                             
        logging.basicConfig(stream=sys.stdout, format='[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s', level=self._log_level)
        self._log = logging.getLogger(__name__)
        
        self._log.warn("RunExperiment Init, perform " + str(expt_config["n_trials"]) + " trials across " + str(expt_config["n_processes"]) + " processes")

        global expt_lock
        expt_lock = Lock()
        
        self._trial_config = []
        self._expt_config = {}

        self.reset(trial_config, expt_config)
        
    def reset(self, trial_config = None, expt_config : dict = None):
        """
        Reset experiment with new configurations

        :param trial_config: (list(dict)/generator) Configurations for each trial, *default*: None
        :param expt_config: (dict) Experiment configuration, *default*: None
        """
        
        if trial_config is not None:
            self._trial_config = trial_config
        if expt_config is not None:
            self._expt_config = nd.merge(self._expt_config, expt_config)

        if "experiment" not in self._expt_config:
            raise ValueError("Must provide experiment function")
        if "n_trials" not in self._expt_config:
            self._expt_config["n_trials"] = 1
        if "n_processes" not in self._expt_config:
            self._expt_config["n_processes"] = 1
        if "save_file" not in self._expt_config:
            self._expt_config["save_file"] = None
        if "clear_save" not in self._expt_config:
            self._expt_config["clear_save"] = False
        if self._expt_config["clear_save"] and self._expt_config["save_file"] is not None and os.path.isfile(self._expt_config["save_file"]):
            self._log.warn("Clearing save file")
            os.remove(self._expt_config["save_file"])

        self._log.warn("Reset experiment")

    def run(self, trial_config = None, expt_config : dict = None):
        """
        Run experiment with new configurations

        :param trial_config: (list(dict)/generator) Configurations for each trial, *default*: None
        :param expt_config: (dict) Experiment configuration, *default*: None
        :return: (pandas.DataFrame) data
        """
        if trial_config is not None or expt_config is not None:
            self.reset(trial_config, expt_config)

        if self._trial_config == []:
            raise ValueError("Must provide trial configuration")
        if self._expt_config["experiment"] == {}:
            raise ValueError("Must provide experiment function")

        self._log.warn("Run experiment")
        self._results = []

        if self._expt_config["n_processes"] == 1:
            return self._run_single()
        else:
            return self._run_multi()

    def _run_single(self):
        """
        Run experiment in single thread

        :return: (pandas.DataFrame) data
        """
        self._log.warn("Run experiment in single thread")
        results = pd.DataFrame()
        for trial in self.__n_iterable(self._trial_config, self._expt_config["n_trials"]):
            result = self._expt_config["experiment"](trial)
            results = pd.concat([results, result])
            if self._expt_config["save_file"] is not None:
                self._save(result)
        return results

    def _run_multi(self):
        """
        Run experiment in multiple processes

        :return: (pandas.DataFrame) data
        """
        self._log.warn("Run experiment in multiple processes")

        results = pd.DataFrame()
        with Pool(self._expt_config["n_processes"]) as p:
            for result in p.imap(self._expt_config["experiment"], self.__n_iterable(self._trial_config, self._expt_config["n_trials"])):
                if self._expt_config["save_file"] is not None:
                    self._save(result) 
                results = pd.concat([results, result])

        return results

    def _save(self, result):
        """
        Save results

        :param result: (pd.DataFrame) result to save
        :return: (pandas.DataFrame) data
        """
        with expt_lock:
            data = import_file(self._expt_config["save_file"])
            data = pd.concat([data, result])
            export_file(data,self._expt_config["save_file"])

    def __n_iterable(self, iterable_el, n = 1):
        """
        Return n copies of an iterable

        :param iterable_el: (iterable) Iterable to copy
        :param n: (int) Number of copies
        :return: () n copies of the input sequences
        """
        for i in range(n):
            if isinstance(iterable_el, list):
                for el in iterable_el:
                    yield el
            else:
                els = deepcopy(iterable_el)
                for el in compile_as_generator(els):
                    yield el


#!/usr/bin/python
"""
Initializes core file data
"""
__license__ = "BSD-3"
__docformat__ = 'reStructuredText'
__author__ = "Jared Beard"

from importlib.resources import path
import sys
import os

current = os.path.dirname(os.path.realpath(__file__))
parent = os.path.dirname(current)
sys.path.append(parent)

import json 

if __name__ == "__main__":
    CORE_FILE_NAME = "/config/core/core.json"
    CORE_DEFAULT_FILE_NAME = "/config/core/core_default.json"

    with open(parent + CORE_DEFAULT_FILE_NAME, 'rb') as f:
        config = json.load(f)

    config["trial_path"] = parent + "/config/"
    config["expt_path"] = parent + "/config/"
    config["save_path"] = parent + "/results/"

    with open(parent + CORE_DEFAULT_FILE_NAME, 'w') as f:
        json.dump(config, f, indent=4)

    with open(parent + CORE_FILE_NAME, 'w') as f:
        json.dump(config, f, indent=4)
        


"""
Functions for importing and exporting files.
"""

__license__ = "BSD-3"
__docformat__ = 'reStructuredText'
__author__ = "Jared Beard"

import sys
import os

current = os.path.dirname(os.path.realpath(__file__))
parent = os.path.dirname(current)
sys.path.append(parent)

import pandas as pd

__all__ = ["import_file", "export_file", "create_file"]

def import_file(filepath : str):
    """
    Import a file as a Pandas dataframe based on its file extension

    :param filepath: (str) file path
    :return: (pandas.DataFrame) the imported dataframe
    """
    extension = filepath.split(".")[-1]
    if not os.path.isfile(filepath):
        return pd.DataFrame()
    if extension == "csv":
        return pd.read_csv(filepath)
    elif extension == "xlsx":
        return pd.read_excel(filepath)
    elif extension == "json":
        return pd.read_json(filepath)
    else:
        raise ValueError(f"Unsupported file type: {extension}")

def export_file(df : pd.DataFrame, filepath : str):
    """
    Write a Pandas dataframe to a file based on its file extension

    :param df: (pandas.DataFrame) the dataframe to be written
    :param filepath: (str) file path
    :return: None
    """
    extension = filepath.split(".")[-1]
    if extension == "csv":
        df.to_csv(filepath, index=False)
    elif extension == "xlsx":
        df.to_excel(filepath, index=False)
    elif extension == "json":
        df.to_json(filepath, index=False)
    else:
        raise ValueError(f"Unsupported file type: {extension}")

def create_file(file_path : str):
    """
    Checks for the existence of file and creates an empty frame if not

    :param file_path: (str) Desired file
    :return: (pd.DataFrame) Data
    """
    file_extension = os.path.splitext(file_path)[1]
    if not os.path.exists(file_path):
        empty_df = pd.DataFrame()
        if file_extension == '.csv':
            empty_df.to_csv(file_path, index=False)
        elif file_extension == '.json':
            empty_df.to_json(file_path)
        elif file_extension == '.xlsx':
            empty_df.to_excel(file_path, index=False)
        else:
            return "Invalid file type"
        return "Empty file created at: " + file_path
    return "File already exists at: " + file_path


#!/usr/bin/env python3
"""
This script is handle command line argmuents for starting experiments and analyzing data.
"""
__license__ = "BSD-3"
__docformat__ = 'reStructuredText'
__author__ = "Jared Beard"

from importlib.resources import path
import sys
import os

current = os.path.dirname(os.path.realpath(__file__))
parent = os.path.dirname(current)
sys.path.append(parent)

from copy import deepcopy
import argparse
import json
import reconfigurator.reconfigurator as rc
from reconfigurator.compiler import compile_as_generator, compile_to_list

from lab_space.experiment import Experiment

import sys
import importlib.util

CORE_FILE_NAME = "/config/core/core.json"
CORE_DEFAULT_FILE_NAME = "/config/core/core_default.json"

__all__ = ['register_experiment', 'deregister_experiment', 'get_registered_experiment']

def register_experiment(module_name :str, func_name : str = None, func_key_name : str = None, module_path : str = None):
    """
    Registers experiment to be run

    :param module_name: (str) Name of experiment file
    :param func_name: (str) Function to call register
    :param module_path: (str) Path to module
    :param func_key_name: (str) Key to identify function
    """
    
    try: 
        f = call_function(module_name, func_name, module_path)
    except:
        raise ValueError("Function not Found")

    with open(current + CORE_FILE_NAME, 'rb') as f:
        core_config = json.load(f)
        temp_dict = {
                "module_path": module_path,
                "module_name": module_name,
                "function_name": func_name
        }
        if func_key_name is None:
            func_key_name = func_name
        core_config["experiments"].update({func_key_name:temp_dict})
    with open(current + CORE_FILE_NAME, 'w') as f:
        json.dump(core_config, f, indent=4)

def deregister_experiment(experiment : str):
    """
    Deregisters experiment to be run

    :param experiment: (str) Experiment to deregister
    """
    with open(current + CORE_FILE_NAME, 'rb') as f:
        core_config = json.load(f)
        if "experiments" in core_config:
            del core_config["experiments"][experiment]
        else: 
            raise ValueError("No Registered Experiments")
    with open(current + CORE_FILE_NAME, 'w') as f:
        json.dump(core_config, f, indent=4)

def get_registered_experiment(experiment : str):
    """
    Gets all registered experiment

    :param experiment: (module) Experiment module
    :return: (function) Experiment function reference
    """
    with open(current + CORE_FILE_NAME, 'rb') as f:
        core_config = json.load(f)
        if "experiments" in core_config:
            expt_config = core_config["experiments"][experiment]
        else: 
            raise ValueError("No Registered Experiments")
    return call_function(expt_config["module_name"], expt_config["function_name"], expt_config["module_path"])

def call_function(module_name : str, func_name : str, module_path :str = None):
    """
    Calls function from module

    :param module_name: (str) Name of experiment file
    :param func_name: (str) Function to call register
    :param module_path: (str) Path to module
    :return: (function) Experiment function reference
    """
    if module_path is not None:
        spec = importlib.util.spec_from_file_location(module_name, module_path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
    else:
        module = importlib.import_module(module_name)
    func = getattr(module, func_name)
    return func

if __name__=='__main__':

    parser = argparse.ArgumentParser(description='Lab Space Experiment CLI')
    parser.add_argument('-r',   '--run',           action="store_const", const=True,  help='Runs algorithm, if unspecified runs user default')
    
    parser.add_argument('-cr',  '--configure_reset',   action="store_const", const=True,  help='Resets all configuration to factory default')
    parser.add_argument('-cp',  '--configure_path',                 type=str, nargs = 1, help='Configures path of config and data files')
    parser.add_argument('-ctp', '--configure_trial_path',           type=str, nargs = 1, help='Configures path of trial config files')
    parser.add_argument('-ct',  '--configure_trial',                type=str, nargs = 1, help='Configures file name for trial config')
    parser.add_argument('-cep', '--configure_experiment_path',      type=str, nargs = 1, help='Configures path of experiment config files')
    parser.add_argument('-ce',  '--configure_experiment',           type=str, nargs = 1, help='Configures file name for experiment config')
    parser.add_argument('-csp', '--configure_save_path',            type=str, nargs = 1, help='Configures path of save files')
    
    parser.add_argument('-tt',  '--num_trials',                     type=int, nargs = 1,  help='Number of trials to run')
    parser.add_argument('-tp',  '--num_processes',                  type=int, nargs = 1,  help='Number of processes to run')
    parser.add_argument('-ts',  '--save_file',                      type=str, nargs ="+", help='Updates file name for save files, if "none" will set to NoneType')
    parser.add_argument('-tcs', '--clear_save',                     type=int ,nargs = 1,  help='Clears save file, 0 -> false, 1 -> true')
    parser.add_argument('-tl',  '--log_level',                      type=str, nargs = 1,  help='Sets log level')
    parser.add_argument('-tc',  '--compile',           action="store_const", const=True,  help='Compiles trial config file')

    parser.add_argument('-e',    '--experiment',                    type=str, nargs = 1,  help='Function to run')
    parser.add_argument('-er',   '--experiment_register',           type=str, nargs ="+", help='Function to register, takes 4 arguments: path, module, name, and function. If only 3 specified, module assumed to be installed/importable')
    parser.add_argument('-ed',   '--experiment_deregister',         type=str, nargs = 1,  help='Deletes experiment from registry')
    
    parser.add_argument('-s',    '--save',             action="store_const", const=True,  help='Saves settings for experiment and trial data to current files. If trial is compiled, will append "c_" to file name and save compiled version')
    parser.add_argument('-st',   '--save_trial',                    type=str, nargs ="+", help='Saves settings for trial data, if argument specified saves to that file in path. If trial is compiled, will append "c_" to file name and save compiled version')
    parser.add_argument('-se',   '--save_experiment',               type=str, nargs ="+", help='Saves settings for experiment data, if argument specified saves to that file in path')
    
    parser.add_argument('-p',    '--print',            action="store_const", const=True,  help='Prints config file')

    args = parser.parse_args()
    #########################################################################################
    # Configurations ------------------------------------------------------------------------
    with open(current + CORE_FILE_NAME, "rb") as f:
        core_config = json.load(f)

    if args.experiment_deregister is not None:
        deregister_experiment(args.experiment_deregister[0])
        exit("Experiment " + args.experiment_deregister[0] + " deregistered")

    with open(current + CORE_DEFAULT_FILE_NAME, "rb") as f:
        core_default = json.load(f)

    if args.configure_reset is not None:
        with open(current + CORE_FILE_NAME, "w") as f:
            json.dump(core_default, f, indent=4)
        print("Core Reset")
    
    if args.configure_path is not None:
        core_config["trial_path"] = args.configure_path[0]
        core_config["expt_path"] = args.configure_path[0]
        core_config["save_path"] = args.configure_path[0]
        rc.write_file(current + CORE_FILE_NAME, core_config)

    if args.configure_trial_path is not None:
        core_config["trial_path"] = args.configure_trial_path[0]
        rc.write_file(current + CORE_FILE_NAME, core_config)
    if args.configure_trial is not None:
        core_config["trial_name"] = args.configure_trial[0]
        rc.write_file(current + CORE_FILE_NAME, core_config)
    trial_config = rc.read_file(core_config["trial_path"] + core_config["trial_name"])

    if args.configure_experiment_path is not None:
        core_config["expt_path"] = args.configure_experiment_path[0]
    if args.configure_experiment is not None:
        core_config["expt_name"] = args.configure_experiment[0]
    expt_config = rc.read_file(core_config["expt_path"] + core_config["expt_name"])

    if args.configure_save_path is not None:
        core_config["save_path"] = args.configure_save_path[0]
        rc.write_file(current + CORE_FILE_NAME, core_config)
    if args.save_file is not None:
        if args.save_file[0] == "none":
            core_config["save_file"] = None
        else:
            expt_config["save_file"] = args.save_file[0]

    if args.num_trials is not None:
        expt_config["n_trials"] = args.num_trials[0]
    if args.num_processes is not None:
        expt_config["n_processes"] = args.num_processes[0]
    if args.clear_save is not None:
        expt_config["clear_save"] = bool(args.clear_save[0])
    if args.log_level is not None:
        expt_config["log_level"] = args.log_level[0]
    
    uncompiled_trial_config = deepcopy(trial_config)
    if args.compile is not None:
        trial_config = compile_as_generator(trial_config)
        

    # Function Registration -----------------------------------------------------------------
    if args.experiment is not None:
        expt_config["experiment"] = args.experiment[0]

    if args.experiment_register is not None:
        if len(args.experiment_register) == 4:
            register_experiment(args.experiment_register[1], args.experiment_register[3], args.experiment_register[2], args.experiment_register[0])
        else:
            register_experiment(args.experiment_register[0], args.experiment_register[2], args.experiment_register[1])

    # Save --------------------------------------------------------------------------------------------
    if args.save is not None and args.save:
        if args.compile is not None:
            temp_tc = compile_to_list(uncompiled_trial_config)
            rc.write_file(core_config["trial_path"] + "c_" + core_config["trial_name"], temp_tc)
        else:
            rc.write_file(core_config["trial_path"] + core_config["trial_name"], trial_config)
        rc.write_file(core_config["expt_path"] + core_config["expt_name"], expt_config)
    
    if args.save_trial is not None:
        if args.save_trial == "none":
            if args.compile is not None:
                temp_tc = compile_to_list(uncompiled_trial_config)
                rc.write_file(core_config["trial_path"] + "c_" + core_config["trial_name"], temp_tc)
            else:
                rc.write_file(core_config["trial_path"] + core_config["trial_name"], trial_config)
        else:
            if not isinstance(trial_config, list):
                temp_tc = compile_to_list(uncompiled_trial_config)
                rc.write_file(core_config["trial_path"] + "c_" + args.save_trial[0], temp_tc)
            else:
                rc.write_file(core_config["trial_path"] + args.save_trial[0], trial_config)
    
    if args.save_experiment is not None:
        if args.save_experiment == "none":
            rc.write_file(core_config["expt_path"] + core_config["expt_name"], expt_config)
        else:
            rc.write_file(core_config["expt_path"] + args.save_experiment[0], expt_config)

    # Print --------------------------------------------------------------------------------------------
    if args.print is not None and args.print:
        print(f'{"Core Config":-<20}')
        rc.print_config(core_config)
        print()
        print()
        print(f'{"Trial Config":-<20}')
        if isinstance(trial_config,dict):
            rc.print_config(trial_config)
        elif args.save is not None and args.save or args.save_trial is not None:
            print(temp_tc)
        else:
            print(uncompiled_trial_config)
        print()
        print()
        print(f'{"Experiment Config":-<20}')
        rc.print_config(expt_config)

    # Run --------------------------------------------------------------------------------------------
    if args.run is not None and args.run:
        expt_config["experiment"] = get_registered_experiment(expt_config["experiment"])
        if expt_config["save_file"] is not None:
            expt_config["save_file"] = core_config["save_path"] + expt_config["save_file"]
        expt = Experiment(trial_config, expt_config, expt_config["log_level"])
        print(expt.run())